'use strict';

import { access } from 'fs/promises';
import { extname } from 'path';
import { Shell } from '../shell';
import { createLog } from '../log';
import type { Logger } from 'winston';

/** @module lib/ffprobe */


interface VideoInfo {
	width : number;
	height : number;
	fps? : number;
	seconds? : number;
	streams? : any[]; //flesh out if needed
	format? : any;
}

/**
 * Class representing all ffprobe features.
 */

export class FFPROBE {
	private bin : string;
	private log : Logger;

	constructor (bin : string = 'ffprobe') {
		this.bin = bin;
		this.log = createLog('ffprobe2');
	}

	public async exists (path : string) : Promise<boolean> {  
		try {
			await access(path);
			return true;
		} catch {
			return false;
		}
	}

	/**
	 * Parse the fps entry into a float representing the fps of a video
	 **/
	private parseFps (fpsStr : string) : number {
		let fps : number = 30.0;
		let parts : string[];
		if (fpsStr.indexOf('/') !== -1) {
			parts = fpsStr.split('/');
			fps = parseFloat(parts[0]) / parseFloat(parts[1]);
		} else {
			fps = parseFloat(fpsStr);
		}
		return fps;
	}
	/**
	 * Get info on a video in json format. Use for filmout.
	 *
	 * @param {string} video Path to video
	 *
	 * @returns {object} Video info in an object
	 **/
	public async info (video : string) : Promise<VideoInfo> {
		const cmd : string[] = [
			this.bin,
			'-v', 'quiet', 
			'-print_format', 'json',
			'-show_format', 
			'-show_streams',
			 video
		];
		let fileExists : boolean;
		let raw : any;
		let json : VideoInfo;
		let vid : any; //whether video has stream with video data

		try {
			fileExists = await this.exists(video);
		} catch (err) {
			this.log.error(`Error checking if ${video} exists`, err);
		}
		if (!fileExists) {
			this.log.error(new Error(`File ${video} does not exist`));
			return null
		}
		
		try {
			raw = await this.exec(cmd);
		} catch (err) {
			this.log.error('Error getting info', err);
			return null
		}

		try {
			json = JSON.parse(raw) as VideoInfo;
		} catch (err) {
			this.log.error('Error parsing stdout', err);
			this.log.error(raw);
			return raw;
		}

		if (json.format && json.format.duration) {
			json.seconds = parseFloat(json.format.duration);
		}

		if (json && json.streams) {
			vid = json.streams.find((stream : any) => {
				if (stream.width && stream.height) return stream;
			});
		}

		if (vid) {
			json.width = vid.width;
			json.height = vid.height;
			json.fps = this.parseFps(vid.r_frame_rate)
		}

		return json;
	}

	private async exec (cmd : string[]) : Promise<string> {
		return new Promise(async (resolve : Function, reject : Function) => {
			function after (lines : string) {
				return resolve(lines);
			}
			const child : Shell = new Shell(cmd, null, null, after, true);
			await child.execute();
		});
		
	}
	/**
	 * Count the number of frames in the video using one of two methods.
	 * The first uses -select_streams and is very fast. The second uses
	 * -count_frames and is VERY slow.
	 *
	 * @param {string} video Path to video
	 *
	 * @returns {integer} Number of frames in video
	 **/
	public async frames (video : string) {
		const ext : string = extname(video.toLowerCase());
		let cmd : string[] = [
			this.bin,
			'-v', 'error', 
			'-select_streams', 'v:0',
			'-show_entries', 'stream=nb_frames',
			'-of', 'default=nokey=1:noprint_wrappers=1',
			video
		];
		let backup_cmd : string[] = [
			this.bin,
			'-v', 'error',
			'-count_frames',
			'-select_streams', 'v:0',
			'-show_entries', 'stream=nb_read_frames',
			'-of', 'default=nokey=1:noprint_wrappers=1',
			video
		];
		let fileExists : boolean;
		let raw : any;
		let frames : number;

		try {
			fileExists = await this.exists(video);
		} catch (err) {
			this.log.error('Error checking if file exists', err);
			return false
		}
		if (!fileExists) {
			this.log.error(new Error(`File ${video} does not exist`));
			return false;
		}
		
		if (ext === '.mkv') {
			cmd = backup_cmd;
		}

		try {
			raw = await this.exec(cmd);
		} catch (err) {
			this.log.error(err);
			return false;
		}

		try {
			frames = parseInt(raw.stdout);
		} catch (err) {
			return raw.stdout;
		}

		return frames;
	}
}

/*
function map (obj : any) {
	console.dir(obj);
}
*/

module.exports = { FFPROBE };
export type { VideoInfo };